"""Parts module.

This module allows PyEnSight to control the parts in the EnSight session.

Example for selecting all 3D parts:

(PyEnSight)
>>> from ansys.pyensight.core import LocalLauncher
>>> session = LocalLauncher().start()
>>> parts = session.ensight.utils.parts
>>> parts.select_by_dimension(3)

(EnSight)
>>> from ensight.utils import parts
>>> parts.select_by_dimension(3)

"""
from types import ModuleType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

try:
    import ensight
    from ensight.objs import ens_emitterobj, ensobjlist  # type: ignore
except ImportError:
    from ansys.api.pyensight.ens_emitterobj import ens_emitterobj
    from ansys.pyensight.core.listobj import ensobjlist

if TYPE_CHECKING:
    from ansys.api.pyensight import ensight_api
    from ansys.api.pyensight.ens_part import ENS_PART
    from ansys.api.pyensight.ens_part_particle_trace import ENS_PART_PARTICLE_TRACE
    from ansys.api.pyensight.ens_var import ENS_VAR


def convert_part(
    _ensight: Union["ensight_api.ensight", "ensight"], part: Union[str, int, "ENS_PART"]
):
    if isinstance(part, str):
        return _ensight.objs.core.PARTS[part][0].PARTNUMBER
    elif isinstance(part, int):
        return part
    elif hasattr(part, "PARTNUMBER"):
        return part.PARTNUMBER


def convert_variable(
    _ensight: Union["ensight_api.ensight", "ensight"], var: Union[str, int, "ENS_VAR"]
) -> Optional[int]:
    if isinstance(var, str):
        return int(_ensight.objs.core.VARIABLES[var][0].ID)
    elif isinstance(var, int):
        return var
    elif hasattr(var, "ID"):
        return int(var.ID)
    return None  # pragma: no cover


class Parts:
    """Controls the parts in the current EnSight ``Session`` instance."""

    class _EnSEmitterPoint(ens_emitterobj):  # pragma: no cover
        def __init__(  # pragma: no cover
            self,
            ensight: "ensight",
            point1: Optional[List[float]] = [0, 0, 0],
        ):  # pragma: no cover
            if not isinstance(ensight, ModuleType):
                raise RuntimeError(
                    "The class cannot be used directly in PyEnSight. It should not be used directly even in EnSight"
                )
            super().__init__(ensight.objs.EMIT_CURSOR)
            self.ensight = ensight
            self.ensight.view_transf.cursor(*point1)
            self.CENTROID = point1

    class _EnSEmitterGrid(ens_emitterobj):  # pragma: no cover
        def __init__(  # pragma: no cover
            self,
            ensight: "ensight",
            point1: Optional[List[float]] = [0, 0, 0],
            point2: Optional[List[float]] = [0, 0, 0],
            point3: Optional[List[float]] = [0, 0, 0],
            point4: Optional[List[float]] = [0, 0, 0],
            num_points_x: Optional[int] = 25,
            num_points_y: Optional[int] = 25,
        ):  # pragma: no cover
            if not isinstance(ensight, ModuleType):
                raise RuntimeError(
                    "The class cannot be used directly in PyEnSight. It should not be used directly even in EnSight"
                )
            super().__init__(ensight.objs.EMIT_PLANE)
            self.ensight = ensight
            self.ensight.view_transf.plane(1, *point1)
            self.ensight.view_transf.plane(2, *point2)
            self.ensight.view_transf.plane(3, *point3)
            self.POINT1 = point1
            self.POINT2 = point2
            self.POINT3 = point3
            self.POINT4 = point4
            self.NUM_POINTS_X = num_points_x
            self.NUM_POINTS_Y = num_points_y

    class _EnSEmitterLine(ens_emitterobj):  # pragma: no cover
        def __init__(  # pragma: no cover
            self,
            ensight: "ensight",
            point1: Optional[List[float]] = [0, 0, 0],
            point2: Optional[List[float]] = [0, 0, 0],
            num_points: Optional[int] = 100,
        ):  # pragma: no cover
            if not isinstance(ensight, ModuleType):
                raise RuntimeError(
                    "The class cannot be used directly in PyEnSight. It should not be used directly even in EnSight"
                )
            super().__init__(ensight.objs.EMIT_LINE)
            self.ensight = ensight
            self.ensight.view_transf.line(1, *point1)
            self.ensight.view_transf.line(2, *point2)
            self.POINT1 = point1
            self.POINT2 = point2
            self.NUM_POINTS = num_points

    class _EnSEmitterPart(ens_emitterobj):  # pragma: no cover
        def __init__(  # pragma: no cover
            self,
            ensight: "ensight",
            part: Optional[Any] = None,
            part_kind: Optional[Any] = 0,
            num_points: Optional[int] = 100,
        ):  # pragma: no cover
            if not isinstance(ensight, ModuleType):
                raise RuntimeError(
                    "The class cannot be used directly in PyEnSight. It should not be used directly even in EnSight"
                )
            super().__init__(ensight.objs.EMIT_PART)
            self.ensight = ensight
            if not part:
                raise RuntimeError("part is a required input")
            self.PART = convert_part(self.ensight, part)
            self.NUM_POINTS = num_points
            self.DISTRIB_TYPE = part_kind

    def __init__(self, ensight: Union["ensight_api.ensight", "ensight"]):
        self.ensight = ensight

    def select_parts_by_dimension(self, dimension: int) -> ensobjlist["ENS_PART"]:
        """Select parts by the input dimension and return the parts found.

        Parameters
        ----------
        dimension : int
            Dimension for selecting parts.

        Returns
        -------
        ensobjlist["ENS_PART"]
            found (ensobjlist): List of parts found.

        """
        parts = self.ensight.objs.core.PARTS
        parts.set_attr("SELECTED", False)
        found = parts.find(True, f"HAS{dimension}DELEMENTS")
        found.set_attr("SELECTED", True)
        return found

    def select_parts_invert(self) -> ensobjlist["ENS_PART"]:
        """Select parts currently not selected, deselecting the previously selected parts.

        Returns
        -------
        ensobjlist["ENS_PART"]
            Updated list of parts selected.

        """
        self.ensight.part.select_invert()
        parts = self.ensight.objs.core.PARTS
        return parts.find(True, "SELECTED")

    def select_parts_by_tag(
        self,
        tag: Optional[str] = None,
        value: Optional[str] = None,
        tagdict: Optional[Dict[str, str]] = None,
    ) -> ensobjlist["ENS_PART"]:
        """Select parts by the input dimension and return the parts found.

        Parameters
        ----------
        tag : str, optional
            Tag for finding the parts.
        value : str, optional
            Value for finding the parts.
        tagdict : dict, optional
            Dictionary containing the key and value pairs for finding
            the parts. Only the parts that have all the keys and corresponding
            values are returned. If a value for this parameter is supplied, it
            takes precedence over the values supplied for the ``tag`` and
            ``value`` parameters.

        Returns
        -------
        ensobjlist["ENS_PART"]
            List of parts found. If no arguments are given, all parts are returned.

        """
        parts = self.ensight.objs.core.PARTS
        metadata = {p: p.METADATA for p in parts}
        found = ensobjlist()
        if not tag and not value and not tagdict:
            self.ensight.part.select_all()
            return parts
        if not tagdict:
            if tag and value:
                found = ensobjlist([p for p, met in metadata.items() if met.get(tag) == value])
            elif value and not tag:
                found = ensobjlist([p for p, met in metadata.items() if value in met.values()])
            elif tag and not value:  # pragma: no cover
                found = ensobjlist([p for p, met in metadata.items() if tag in met.keys()])
        else:
            found = ensobjlist(
                [
                    p
                    for p, met in metadata.items()
                    if all(met.get(k) == v for k, v in tagdict.items())
                ]
            )
        if found:
            found.set_attr("SELECTED", True)
        return found

    _EMIT_POINT: int = 0
    _EMIT_LINE: int = 1
    _EMIT_PLANE: int = 2
    _EMIT_PART: int = 3
    PT_POS_TIME: str = "+"
    PT_NEG_TIME: str = "-"
    PT_POS_NEG_TIME: str = "+/-"
    PART_EMIT_FROM_NODES: int = 0
    PART_EMIT_FROM_AREA: int = 1

    def _create_emitters(
        self,
        emitter_type: int,
        points: Optional[List[List[float]]] = None,
        point1: Optional[List[float]] = None,
        point2: Optional[List[float]] = None,
        point3: Optional[List[float]] = None,
        parts: Optional[List["ENS_PART"]] = None,
        part_distribution_type: Optional[int] = 0,
        num_points: Optional[int] = 100,
        num_points_x: Optional[int] = 25,
        num_points_y: Optional[int] = 25,
    ) -> List[Any]:
        """Private routine to create emitter objects"""
        new_emitters: List[Any] = []
        if emitter_type == self._EMIT_POINT:
            if not points:  # pragma: no cover
                raise RuntimeError(
                    "list of points needed if particle trace emitted from points"
                )  # pragma: no cover
            for p in points:
                if isinstance(self.ensight, ModuleType):  # pragma: no cover
                    new_emitters.append(
                        self._EnSEmitterPoint(self.ensight, point1=p)
                    )  # pragma: no cover
                else:
                    new_emitters.append(
                        f"ensight.utils.parts._EnSEmitterPoint(ensight, point1={p})"
                    )
        elif emitter_type == self._EMIT_LINE:
            if not any([point1, point2]):
                raise RuntimeError("point1 and point2 needed if particle trace emitted from line")
            if isinstance(self.ensight, ModuleType):  # pragma: no cover
                new_emitters.append(  # pragma: no cover
                    self._EnSEmitterLine(
                        self.ensight, point1=point1, point2=point2, num_points=num_points
                    )
                )
            else:
                new_emitters.append(
                    f"ensight.utils.parts._EnSEmitterLine(ensight, point1={point1}, point2={point2}, num_points={num_points})"
                )
        elif emitter_type == self._EMIT_PLANE:
            if not any([point1, point2, point3]):
                raise RuntimeError(  # pragma: no cover
                    "point1, point2 and point3 needed if particle trace emitted from plane"
                )
            if isinstance(self.ensight, ModuleType):  # pragma: no cover
                new_emitters.append(  # pragma: no cover
                    self._EnSEmitterGrid(
                        self.ensight,
                        point1=point1,
                        point2=point2,
                        point3=point3,
                        num_points_x=num_points_x,
                        num_points_y=num_points_y,
                    )
                )
            else:
                new_emitters.append(
                    f"ensight.utils.parts._EnSEmitterGrid(ensight, point1={point1}, point2={point2}, point3={point3}, num_points_x={num_points_x}, num_points_y={num_points_y})"
                )
        elif emitter_type == self._EMIT_PART:  # pragma: no cover
            if not parts:  # pragma: no cover
                raise RuntimeError(
                    "part and num_points needed if particle trace emitted from part"
                )  # pragma: no cover
            for p in parts:
                if isinstance(self.ensight, ModuleType):  # pragma: no cover
                    new_emitters.append(  # pragma: no cover
                        self._EnSEmitterPart(
                            self.ensight,
                            part=p,
                            num_points=num_points,
                            part_kind=part_distribution_type,
                        )
                    )
                else:
                    new_emitters.append(
                        f"ensight.utils.parts._EnSEmitterPart(ensight, part={convert_part(self.ensight ,p)}, num_points={num_points}, part_kind={part_distribution_type})"
                    )
        else:
            raise RuntimeError(
                "No input provided to create the emitters for the particle trace"
            )  # pragma: no cover
        return new_emitters

    def _create_particle_trace_part(
        self,
        name: str,
        variable: Union[str, int, "ENS_VAR"],
        direction: str,
        source_parts: List["ENS_PART"],
        pathlines: Optional[bool] = False,
        emit_time: Optional[float] = None,
        total_time: Optional[float] = None,
        delta_time: Optional[float] = None,
        surface_restrict: Optional[bool] = False,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """Private routine to create a particle trace part object"""
        current_timestep = None
        direction_map = {
            self.PT_POS_TIME: self.ensight.objs.enums.POS_TIME,
            self.PT_NEG_TIME: self.ensight.objs.enums.NEG_TIME,
            self.PT_POS_NEG_TIME: self.ensight.objs.enums.POS_NEG_TIME,
        }
        idx = self.ensight.objs.enums.PART_PARTICLE_TRACE
        def_part: "ENS_PART_PARTICLE_TRACE" = self.ensight.objs.core.DEFAULTPARTS[idx]
        def_part.TYPE = self.ensight.objs.enums.STREAMLINE
        if pathlines is True:
            def_part.TYPE = self.ensight.objs.enums.PATHLINE
            current_timestep = self.ensight.objs.core.TIMESTEP
            self.ensight.objs.core.TIMESTEP = self.ensight.objs.core.TIMESTEP_LIMITS[0]
        if total_time:
            def_part.TOTALTIME = total_time
        if delta_time:
            def_part.DELTATIME = delta_time
        if emit_time:  # pragma: no cover
            def_part.STARTTIME = emit_time
        def_part.DESCRIPTION = name
        def_part.VARIABLE = convert_variable(self.ensight, variable)
        def_part.SURFACERESTRICTED = False
        def_part.TRACEDIRECTION = direction_map.get(direction)
        if surface_restrict:
            def_part.SURFACERESTRICTED = True
        particle_trace_part: "ENS_PART_PARTICLE_TRACE" = def_part.createpart(
            sources=source_parts, name=name
        )[0]
        if current_timestep:
            self.ensight.objs.core.TIMESTEP = current_timestep
        return particle_trace_part

    def _add_emitters_to_particle_trace_part(
        self,
        particle_trace_part: "ENS_PART_PARTICLE_TRACE",
        new_emitters: List[Any],
        palette: Optional[str] = None,
        clean: Optional[bool] = False,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """Private utility to add emitters to an existing particle trace part."""
        if isinstance(self.ensight, ModuleType):  # pragma: no cover
            if clean:  # pragma: no cover
                emitters = []  # pragma: no cover
            else:  # pragma: no cover
                emitters = particle_trace_part.EMITTERS.copy()  # pragma: no cover
            emitters.extend(new_emitters)  # pragma: no cover
            particle_trace_part.EMITTERS = emitters  # pragma: no cover
        else:
            if clean:
                self.ensight._session.cmd("enscl.emitters=[]", do_eval=False)
            else:
                self.ensight._session.cmd(
                    f"enscl.emitters=ensight.objs.wrap_id({particle_trace_part.objid}).EMITTERS.copy()",
                    do_eval=False,
                )
            text = "enscl.emitters.extend(["
            for emitter in new_emitters:
                text += emitter + ", "
            text = text[:-2]
            text += "])"
            self.ensight._session.cmd(text, do_eval=False)
            self.ensight._session.cmd(
                f"ensight.objs.wrap_id({particle_trace_part.objid}).setattr('EMITTERS', enscl.emitters.copy())"
            )
            self.ensight._session.cmd("del enscl.emitters", do_eval=False)
        if palette:
            particle_trace_part.COLORBYPALETTE = palette
        return particle_trace_part

    def _cure_particle_trace_part(
        self, particle_trace_part: Union[str, int, "ENS_PART_PARTICLE_TRACE"]
    ) -> "ENS_PART_PARTICLE_TRACE":
        """Private utility to cure an input particle trace part and convert it to an ``ENS_PART`"""

        # the add_emitter* functions were added in 2024 R2
        if not isinstance(self.ensight, ModuleType):  # pragma: no cover
            self.ensight._session.ensight_version_check("2024 R2")

        _particle_trace_part: "ENS_PART_PARTICLE_TRACE"
        if isinstance(particle_trace_part, (str, int)):  # pragma: no cover
            temp = self.ensight.objs.core.PARTS[particle_trace_part]  # pragma: no cover
            if not temp:  # pragma: no cover
                raise RuntimeError(
                    "particle_trace_part input is not a valid part"
                )  # pragma: no cover
            _particle_trace_part = temp[0]  # pragma: no cover
        else:
            _particle_trace_part = particle_trace_part
        return _particle_trace_part

    def _prepare_particle_creation(
        self,
        direction: Optional[str] = None,
        source_parts: Optional[List[Union[str, int, "ENS_PART"]]] = None,
    ) -> Tuple[str, List["ENS_PART"]]:
        """Private utility to set the direction if not provided, and to cure the list of source parts."""

        # the create_particle* functions were added in 2024 R2
        if not isinstance(self.ensight, ModuleType):  # pragma: no cover
            self.ensight._session.ensight_version_check("2024 R2")

        if not direction:
            direction = self.PT_POS_TIME
        if source_parts:  # pragma: no cover
            converted_source_parts = [convert_part(self.ensight, p) for p in source_parts]
        if not source_parts:  # pragma: no cover
            converted_source_parts = self.ensight.objs.core.selection(  # pragma: no cover
                name="ENS_PART"
            )
        if not converted_source_parts:  # pragma: no cover
            raise RuntimeError("No part selected for particle trace generation")  # pragma: no cover
        return direction, converted_source_parts

    def _find_palette(self, color_by: Optional[Union[str, int, "ENS_VAR"]] = None) -> Optional[str]:
        """Private utility to find the description of the input color_by variable"""
        palette: Optional[str] = None
        if color_by:
            try:
                _color_by_var: List["ENS_VAR"] = self.ensight.objs.core.VARIABLES.find(
                    [convert_variable(self.ensight, color_by)], attr="ID"
                )
                if _color_by_var:
                    palette = _color_by_var[0].DESCRIPTION
            except Exception:
                raise RuntimeError(
                    "The variable supplied to color the particle trace by does not exist"
                )
        return palette

    def create_particle_trace_from_points(
        self,
        name: str,
        variable: Union[str, int, "ENS_VAR"],
        points: List[List[float]],
        direction: Optional[str] = None,
        pathlines: Optional[bool] = False,
        source_parts: Optional[List[Union[str, int, "ENS_PART"]]] = None,
        emit_time: Optional[float] = None,
        total_time: Optional[float] = None,
        delta_time: Optional[float] = None,
        color_by: Optional[Union[str, int, "ENS_VAR"]] = None,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Create a particle trace part from a list o points.
        Returns the ``ENS_PART`` generated.

        Parameters
        ----------

        name: str
            The name of part to be generated
        variable:
            The variable to compute the particle traces with.
            It can be the name, the ID or the ``ENS_VAR`` object. It must be a vector variable.
        direction: str
            The direction for the particle traces to be generated.
            This table describes the options:

            ================== ==============================================
            Name               Query type
            ================== ==============================================
            PT_POS_TIME        Follow the vector direction
            PT_NEG_TIME        Go contrary to the vector direction
            PT_POS_NEG_TIME    Follow and go contrary to the vector direction
            ================== ==============================================

            If not provided, it will default to ``PT_POS_TIME``
        pathlines: bool
            True if the particle traces need to be pathlines
        points: list
            List of coordinates for the seed points.
        source_parts: list
            A list of parts to create the particle trace in. For instance, in a CFD
            simulation this might be the fluid zone.
            If not provided, the function will try to look for the selected parts.
        emit_time: float
            The emission time to start the particle trace from. If not provided,
            it will use the current time.
        total_time: float
            The total emission time. If not provided, EnSight will provide the end time
            for a transient simulation, an internal best time for steady state simulations.
        delta_time: float
            The interval for the emissions. If not provided, EnSight will provide
            a best estimate.
        color_by
            The optional variable to color the particle trace by.
            It can be the name, the ID or the ``ENS_VAR`` object.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> s.ensight.utils.parts.create_particle_trace_from_points("mytraces", "Velocity", points=[[-0.02,-0.123,0.01576],[0.109876,-0.123,0.0123]], source_parts=parts.select_parts_by_dimension(3))
        """
        emitter_type = self._EMIT_POINT
        direction, converted_source_parts = self._prepare_particle_creation(
            direction=direction, source_parts=source_parts
        )
        particle_trace_part = self._create_particle_trace_part(
            name,
            variable,
            direction,
            converted_source_parts,
            pathlines=pathlines,
            emit_time=emit_time,
            delta_time=delta_time,
            total_time=total_time,
        )
        new_emitters = self._create_emitters(emitter_type=emitter_type, points=points)
        palette = self._find_palette(color_by=color_by)
        return self._add_emitters_to_particle_trace_part(
            particle_trace_part, new_emitters=new_emitters, palette=palette, clean=True
        )

    def create_particle_trace_from_line(
        self,
        name: str,
        variable: Union[str, int, "ENS_VAR"],
        point1: List[float],
        point2: List[float],
        num_points: Optional[int] = 100,
        direction: Optional[str] = None,
        pathlines: Optional[bool] = False,
        source_parts: Optional[List[Union[str, int, "ENS_PART"]]] = None,
        emit_time: Optional[float] = None,
        total_time: Optional[float] = None,
        delta_time: Optional[float] = None,
        color_by: Optional[Union[str, int, "ENS_VAR"]] = None,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Create a particle trace part from a line.
        Returns the ``ENS_PART`` generated.

        Parameters
        ----------

        name: str
            The name of part to be generated
        variable:
            The variable to compute the particle traces with.
            It can be the name, the ID or the ``ENS_VAR`` object. It must be a vector variable.
        direction: str
            The direction for the particle traces to be generated.
            This table describes the options:

            ================== ==============================================
            Name               Query type
            ================== ==============================================
            PT_POS_TIME        Follow the vector direction
            PT_NEG_TIME        Go contrary to the vector direction
            PT_POS_NEG_TIME    Follow and go contrary to the vector direction
            ================== ==============================================

            If not provided, it will default to ``PT_POS_TIME``
        pathlines: bool
            True if the particle traces need to be pathlines
        point1: list
            List of coordinates for point 1.
        point2: list
            List of coordinates for point 2.
        source_parts: list
            A list of parts to create the particle trace in. For instance, in a CFD
            simulation this might be the fluid zone.
            If not provided, the function will try to look for the selected parts.
        num_points: int
            The number of points to emit from. Defaults to 100.
        emit_time: float
            The emission time to start the particle trace from. If not provided,
            it will use the current time.
        total_time: float
            The total emission time. If not provided, EnSight will provide the end time
            for a transient simulation, an internal best time for steady state simulations.
        delta_time: float
            The interval for the emissions. If not provided, EnSight will provide
            a best estimate.
        color_by
            The optional variable to color the particle trace by.
            It can be the name, the ID or the ``ENS_VAR`` object.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> parts = s.ensight.utils.parts
        >>> parts.create_particle_trace_from_line("mytraces", "Velocity", point1=[-0.02,-0.123,0.01576], point2=[0.109876,-0.123,0.0123], num_points=10, source_parts=parts.select_parts_by_dimension(3))
        """
        emitter_type = self._EMIT_LINE
        direction, converted_source_parts = self._prepare_particle_creation(
            direction=direction, source_parts=source_parts
        )
        particle_trace_part = self._create_particle_trace_part(
            name,
            variable,
            direction,
            converted_source_parts,
            pathlines=pathlines,
            emit_time=emit_time,
            delta_time=delta_time,
            total_time=total_time,
        )
        new_emitters = self._create_emitters(
            emitter_type=emitter_type, point1=point1, point2=point2, num_points=num_points
        )
        palette = self._find_palette(color_by=color_by)
        return self._add_emitters_to_particle_trace_part(
            particle_trace_part, new_emitters=new_emitters, palette=palette, clean=True
        )

    def create_particle_trace_from_plane(
        self,
        name: str,
        variable: Union[str, int, "ENS_VAR"],
        point1: List[float],
        point2: List[float],
        point3: List[float],
        num_points_x: Optional[int] = 25,
        num_points_y: Optional[int] = 25,
        direction: Optional[str] = None,
        pathlines: Optional[bool] = False,
        source_parts: Optional[List[Union[str, int, "ENS_PART"]]] = None,
        emit_time: Optional[float] = None,
        total_time: Optional[float] = None,
        delta_time: Optional[float] = None,
        color_by: Optional[Union[str, int, "ENS_VAR"]] = None,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Create a particle trace part from a plane.
        Returns the ``ENS_PART`` generated.

        Parameters
        ----------

        name: str
            The name of part to be generated
        variable:
            The variable to compute the particle traces with.
            It can be the name, the ID or the ``ENS_VAR`` object. It must be a vector variable.
        direction: str
            The direction for the particle traces to be generated.
            This table describes the options:

            ================== ==============================================
            Name               Query type
            ================== ==============================================
            PT_POS_TIME        Follow the vector direction
            PT_NEG_TIME        Go contrary to the vector direction
            PT_POS_NEG_TIME    Follow and go contrary to the vector direction
            ================== ==============================================

            If not provided, it will default to ``PT_POS_TIME``
        pathlines: bool
            True if the particle traces need to be pathlines
        point1: list
            List of coordinates for point 1, being a corner of the plane.
        point2: list
            List of coordinates for point 2, being a corner of the plane.
        point3: list
            List of coordinates for point 3, being a corner of the plane.
        source_parts: list
            A list of parts to create the particle trace in. For instance, in a CFD
            simulation this might be the fluid zone.
            If not provided, the function will try to look for the selected parts.
        num_points_x: int
            The number of points on the ``X`` direction of the emission plane.
            Defaults to 25.
        num_points_y: int
            The number of points on the ``Y`` direction of the emission plane.
            Defaults to 25.
        emit_time: float
            The emission time to start the particle trace from. If not provided,
            it will use the current time.
        total_time: float
            The total emission time. If not provided, EnSight will provide the end time
            for a transient simulation, an internal best time for steady state simulations.
        delta_time: float
            The interval for the emissions. If not provided, EnSight will provide
            a best estimate.
        color_by
            The optional variable to color the particle trace by.
            It can be the name, the ID or the ``ENS_VAR`` object.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> parts = s.ensight.utils.parts
        >>> parts.create_particle_trace_from_plane("mytraces", "Velocity", point1=[-0.02,-0.123,0.01576], point2=[0.109876,-0.123,0.0123], point3=[0.1, 0, 0.05] ,num_points_x=10, num_points_y=10, source_parts=parts.select_parts_by_dimension(3))
        """
        emitter_type = self._EMIT_PLANE
        direction, converted_source_parts = self._prepare_particle_creation(
            direction=direction, source_parts=source_parts
        )
        particle_trace_part = self._create_particle_trace_part(
            name,
            variable,
            direction,
            converted_source_parts,
            pathlines=pathlines,
            emit_time=emit_time,
            delta_time=delta_time,
            total_time=total_time,
        )
        new_emitters = self._create_emitters(
            emitter_type=emitter_type,
            point1=point1,
            point2=point2,
            point3=point3,
            num_points_x=num_points_x,
            num_points_y=num_points_y,
        )
        palette = self._find_palette(color_by=color_by)
        return self._add_emitters_to_particle_trace_part(
            particle_trace_part, new_emitters=new_emitters, palette=palette, clean=True
        )

    def create_particle_trace_from_parts(
        self,
        name: str,
        variable: Union[str, int, "ENS_VAR"],
        parts: List[Union[str, int, "ENS_PART"]],
        part_distribution_type: Optional[int] = 0,
        num_points: Optional[int] = 100,
        direction: Optional[str] = None,
        pathlines: Optional[bool] = False,
        source_parts: Optional[List[Union[str, int, "ENS_PART"]]] = None,
        emit_time: Optional[float] = None,
        total_time: Optional[float] = None,
        delta_time: Optional[float] = None,
        color_by: Optional[Union[str, int, "ENS_VAR"]] = None,
        surface_restrict: Optional[bool] = False,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Create a particle trace part from a list of seed parts.
        Returns the ``ENS_PART`` generated.

        Parameters
        ----------

        name: str
            The name of part to be generated
        variable:
            The variable to compute the particle traces with.
            It can be the name, the ID or the ``ENS_VAR`` object. It must be a vector variable.
        direction: str
            The direction for the particle traces to be generated.
            This table describes the options:

            ================== ==============================================
            Name               Query type
            ================== ==============================================
            PT_POS_TIME        Follow the vector direction
            PT_NEG_TIME        Go contrary to the vector direction
            PT_POS_NEG_TIME    Follow and go contrary to the vector direction
            ================== ==============================================

            If not provided, it will default to ``PT_POS_TIME``
        pathlines: bool
            True if the particle traces need to be pathlines
        source_parts: list
            A list of parts to create the particle trace in. For instance, in a CFD
            simulation this might be the fluid zone.
            If not provided, the function will try to look for the selected parts.
        parts: list
            A list of parts to emit the particle traces from.
            They can be their names, their IDs or the respective ``ENS_PART`` objects.
        part_distribution_type: int
            The distribution of emitters in case of emission from a part.
            This table describes the options:

            ====================   =================================================
            Name                   Query type
            ====================   =================================================
            PART_EMIT_FROM_NODES   Emit from the nodes of the part
            PART_EMIT_FROM_AREA    Create an area of equidistant points for emission
            ====================   =================================================

            If not provided, it will default to ``PART_EMIT_FROM_NODES``
        num_points: int
            The number of points to emit from.
            Defaults to 100.
        emit_time: float
            The emission time to start the particle trace from. If not provided,
            it will use the current time.
        total_time: float
            The total emission time. If not provided, EnSight will provide the end time
            for a transient simulation, an internal best time for steady state simulations.
        delta_time: float
            The interval for the emissions. If not provided, EnSight will provide
            a best estimate.
        color_by
            The optional variable to color the particle trace by.
            It can be the name, the ID or the ``ENS_VAR`` object.
        surface_restrict: bool
            True if the particle trace needs to be restricted to the input parts.
            Defaults to False. The flag will be applied to any additional emitter
            appended to the particle trace created.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> parts = s.ensight.utils.parts
        >>> parts.create_particle_trace_from_parts("mytraces", "Velocity", parts=["hot-inlet", "cold-inlet"], num_points=100 source_parts=parts.select_parts_by_dimension(3))
        """
        emitter_type = self._EMIT_PART
        direction, converted_source_parts = self._prepare_particle_creation(
            direction=direction, source_parts=source_parts
        )
        particle_trace_part = self._create_particle_trace_part(
            name,
            variable,
            direction,
            converted_source_parts,
            pathlines=pathlines,
            emit_time=emit_time,
            delta_time=delta_time,
            total_time=total_time,
            surface_restrict=surface_restrict,
        )
        new_parts = [convert_part(self.ensight, p) for p in parts]
        new_emitters = self._create_emitters(
            emitter_type=emitter_type,
            parts=new_parts,
            part_distribution_type=part_distribution_type,
            num_points=num_points,
        )
        palette = self._find_palette(color_by=color_by)
        return self._add_emitters_to_particle_trace_part(
            particle_trace_part, new_emitters=new_emitters, palette=palette, clean=True
        )

    def add_emitter_points_to_particle_trace_part(
        self,
        particle_trace_part: Union[str, int, "ENS_PART"],
        points: List[List[float]],
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Add point emitters to an existing particle trace. The function will return the updated
        ``ENS_PART`` object.

        Parameters
        ----------

        particle_trace_part:
            The particle trace part to be added emitters to.
            Can be the name, the ID or the ``ENS_PART`` object
        points: list
            List of list containing the coordinates for the seed points.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> p = s.ensight.utils.parts.create_particle_trace_from_points("mytraces", "Velocity", points=[[-0.02, -0.123, 0.01576]], source_parts=parts.select_parts_by_dimension(3))
        >>> p = s.ensight.utils.parts.add_emitter_points_to_particle_trace_part(p, points=[[0.109876, -0.123, 0.0123]])
        """
        emitter_type = self._EMIT_POINT
        particle_trace_part = self._cure_particle_trace_part(particle_trace_part)
        new_emitters = self._create_emitters(emitter_type=emitter_type, points=points)
        return self._add_emitters_to_particle_trace_part(particle_trace_part, new_emitters)

    def add_emitter_line_to_particle_trace_part(
        self,
        particle_trace_part: Union[str, int, "ENS_PART"],
        point1: List[float],
        point2: List[float],
        num_points: Optional[int] = 100,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Add a line emitter to an existing particle trace. The function will return the updated
        ``ENS_PART`` object.

        Parameters
        ----------

        particle_trace_part:
            The particle trace part to be added emitters to.
            Can be the name, the ID or the ``ENS_PART`` object.
        point1: list
            The coordinates for point 1.
        point2: list
            The coordinates for point 2.
        num_points: int
            The number of seed points. Defaults to 100.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> p = s.ensight.utils.parts.create_particle_trace_from_points("mytraces", "Velocity", points=[[-0.02,-0.123,0.01576]], source_parts=parts.select_parts_by_dimension(3))
        >>> p = s.ensight.utils.parts.add_emitter_line_to_particle_trace_part(p, point1=[-0.02, -0.123, 0.01576], point2=[0.109876, -0.123, 0.0123], num_points=10)
        """
        emitter_type = self._EMIT_LINE
        particle_trace_part = self._cure_particle_trace_part(particle_trace_part)
        new_emitters = self._create_emitters(
            emitter_type=emitter_type, point1=point1, point2=point2, num_points=num_points
        )
        return self._add_emitters_to_particle_trace_part(particle_trace_part, new_emitters)

    def add_emitter_plane_to_particle_trace_part(
        self,
        particle_trace_part: Union[str, int, "ENS_PART"],
        point1: List[float],
        point2: List[float],
        point3: List[float],
        num_points_x: Optional[int] = 25,
        num_points_y: Optional[int] = 25,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Add a plane emitter to an existing particle trace. The function will return the updated
        ``ENS_PART`` object.

        Parameters
        ----------

        particle_trace_part:
            The particle trace part to be added emitters to.
            Can be the name, the ID or the ``ENS_PART`` object.
        point1: list
            The coordinates for point 1, being a corner of the plane.
        point2: list
            The coordinates for point 2, being a corner of the plane.
        point3: list
            The coordinates for point 3, being a corner of the plane.
        num_points_x: int
            The number of points on the ``X`` direction of the emission plane.
            Defaults to 25.
        num_points_y: int
            The number of points on the ``Y`` direction of the emission plane.
            Defaults to 25.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> p = s.ensight.utils.parts.create_particle_trace_from_points("mytraces", "Velocity", points=[[-0.02,-0.123,0.01576]], source_parts=parts.select_parts_by_dimension(3))
        >>> p = s.ensight.utils.parts.add_emitter_plane_to_particle_trace_part(p, point1=[-0.02, -0.123, 0.01576], point2=[0.109876, -0.123, 0.0123], point3=[0.1, 0, 0.05], num_points_x=10, num_points_y=10)
        """
        emitter_type = self._EMIT_PLANE
        particle_trace_part = self._cure_particle_trace_part(particle_trace_part)
        new_emitters = self._create_emitters(
            emitter_type=emitter_type,
            point1=point1,
            point2=point2,
            point3=point3,
            num_points_x=num_points_x,
            num_points_y=num_points_y,
        )
        return self._add_emitters_to_particle_trace_part(particle_trace_part, new_emitters)

    def add_emitter_parts_to_particle_trace_part(
        self,
        particle_trace_part: Union[str, int, "ENS_PART"],
        parts: List[Union[str, int, "ENS_PART"]],
        part_distribution_type: Optional[int] = 0,
        num_points: Optional[int] = 100,
    ) -> "ENS_PART_PARTICLE_TRACE":
        """
        Add a list of part emitters to an existing particle trace. The function will return the updated
        ``ENS_PART`` object.

        Parameters
        ----------

        particle_trace_part:
            The particle trace part to be added emitters to.
            Can be the name, the ID or the ``ENS_PART`` object.
        parts: list
            A list of parts to emit the particle traces from.
            They can be their names, their IDs or the respective ``ENS_PART`` objects.
        part_distribution_type: int
            The distribution of emitters in case of emission from a part.
            This table describes the options:

            ====================   =================================================
            Name                   Query type
            ====================   =================================================
            PART_EMIT_FROM_NODES   Emit from the nodes of the part
            PART_EMIT_FROM_AREA    Create an area of equidistant points for emission
            ====================   =================================================

            If not provided, it will default to ``PART_EMIT_FROM_NODES``
        num_points: int
            The number of points to emit from.
            Defaults to 100.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> cas_file = s.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = s.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> s.load_data(cas_file, result_file=dat_file)
        >>> p = s.ensight.utils.parts.create_particle_trace_from_points("mytraces", "Velocity", points=[[-0.02, -0.123, 0.01576]], source_parts=parts.select_parts_by_dimension(3))
        >>> p = s.ensight.utils.parts.add_emitter_parts_to_particle_trace_part(p, parts=["cold-inlet", "hot-inlet"], num_points=25)
        """
        emitter_type = self._EMIT_PART
        particle_trace_part = self._cure_particle_trace_part(particle_trace_part)
        new_parts = [convert_part(self.ensight, p) for p in parts]
        new_emitters = self._create_emitters(
            emitter_type=emitter_type,
            parts=new_parts,
            part_distribution_type=part_distribution_type,
            num_points=num_points,
        )
        return self._add_emitters_to_particle_trace_part(particle_trace_part, new_emitters)

    def select_parts(
        self,
        p_list: Optional[List[Union[str, int, "ENS_PART"]]] = None,
        rec_flag: Optional[bool] = True,
    ) -> Optional[List["ENS_PART"]]:
        """
        Select the parts string, or int, or ensight.objs.ENS_PART, or list
        and record the selection (by default) honoring the
        EnSight preference to record command language by part id or by name.
        It creates a list of part objects and selects the parts, and records the
        selection by default.

        Parameters
        ----------
        p_list: list
            The list of part objects to compute the forces on. It can either be a list of names
            a list of IDs (integers or strings) or directly a list of ENS_PART objects
        rec_flag: bool
            True if the selection needs to be recorded

        Returns
        -------
        list
            A list of part objects selected or None if error.


        NOTE: If you do not want a measured part in your
                selection, then don't include it in the list
                e.g. if
                core.PARTS[0].PARTTYPE == ensight.objs.enums.PART_DISCRETE_PARTICLE == 3
                then it is a measured part
        """
        #
        pobj_list = self.get_part_id_obj_name(p_list, "obj")

        if not pobj_list:
            raise RuntimeError("Error, select_parts: part list is empty")
        else:
            # This was formerly used to record command lang 10.1.6(c)
            #  using part ids:
            #  ensight.part.select_begin(pid_list,record=1)
            # Now records selection, honoring the the preference
            #   part selection of by part id or by name (2024R1)
            record = 1 if rec_flag else 0
            self.ensight.objs.core.selection(name="ENS_PART").addchild(
                pobj_list, replace=1, record=record
            )
            # This is essential to synchronize cmd lang with the GUI, C++
            self.ensight.part.get_mainpartlist_select()

        return pobj_list

    def get_part_id_obj_name(
        self,
        plist: Optional[Union[str, int, "ENS_PART", List[str], List[int], List["ENS_PART"]]] = None,
        ret_flag="id",
    ) -> Union[Optional[List[int]], Optional[List[str]], Optional[List["ENS_PART"]]]:
        """
        Input a part or a list of parts and return an id, object, or name
        or a list of ids, objects, or names.

        Parameters
        ----------
        p_list: list
            The list of part objects to compute the forces on. It can either be a list of names
            a list of IDs (integers or strings) or directly a list of ENS_PART objects

        ret_flag: str
            A string that determines what is returned

        Returns
        -------
        list
            Either a list of part IDs, or a list of names or a list of ENS_PART objects
            depending on the requested ret_flag value
        """
        # To not change the interface I didn't move ret_flag to be a required argument,
        # so I need to check its value now
        if not ret_flag:
            return None
        if not plist:
            plist = [p for p in self.ensight.objs.core.PARTS]
        pobj_list: List["ENS_PART"] = []
        #
        #  Basically figure out what plist is, then convert it to a list of ENS_PARTs
        #
        if (
            isinstance(plist, self.ensight.objs.ENS_PART)
            or isinstance(plist, int)
            or isinstance(plist, str)
        ):
            p_list = [plist]
        elif isinstance(plist, list) or isinstance(plist, ensobjlist):
            p_list = [p for p in plist]
        else:  # pragma: no cover
            raise RuntimeError(  # pragma: no cover
                "Unknown type of input var plist {}".format(type(plist))
            )
        #
        #  p_list must now be a list
        #

        if not p_list:
            return None
        if not isinstance(p_list[0], (str, int, self.ensight.objs.ENS_PART)):  # pragma: no cover
            error = "First member is neither ENS_PART, int, nor string"  # pragma: no cover
            error += f"{p_list[0]} type = {type(p_list[0])}; aborting"  # pragma: no cover
            raise RuntimeError(error)  # pragma: no cover
        pobjs: List["ENS_PART"]
        if isinstance(p_list[0], int):
            # list of ints must be part ids
            for pid in p_list:
                pobjs = [p for p in self.ensight.objs.core.PARTS if p.PARTNUMBER == pid]
                for prt in pobjs:
                    pobj_list.append(prt)
        elif isinstance(p_list[0], str):
            if not p_list[0].isdigit():
                for pname in p_list:
                    pobjs = [p for p in self.ensight.objs.core.PARTS if p.DESCRIPTION == pname]
                    for prt in pobjs:
                        pobj_list.append(prt)
            else:  # digits, must be a string list of part ids?
                for pid_str in p_list:
                    pobjs = [
                        p for p in self.ensight.objs.core.PARTS if p.PARTNUMBER == int(pid_str)
                    ]
                    for prt in pobjs:
                        pobj_list.append(prt)
        else:
            for prt in p_list:
                pobj_list.append(prt)
        if ret_flag == "name":
            val_strings = [str(p.DESCRIPTION) for p in pobj_list]
            return val_strings
        if ret_flag == "obj":
            val_objs = [p for p in pobj_list]
            return val_objs
        val_ints = [int(p.PARTNUMBER) for p in pobj_list]
        return val_ints



import glob
import json
import os
import platform
import subprocess
import sys
import tempfile
import threading
from types import ModuleType
from typing import TYPE_CHECKING, List, Optional, Union
from urllib.parse import ParseResult, urlparse
import uuid

import psutil

if TYPE_CHECKING:
    try:
        import ensight
    except ImportError:
        from ansys.api.pyensight import ensight_api


def _handle_fluids_one(install_path):
    cei_path = install_path
    interpreter = os.path.join(cei_path, "bin", "cpython")
    if platform.system() == "Windows":
        interpreter += ".bat"
    return interpreter


class OmniverseKitInstance:
    """Interface to an Omniverse application instance

    Parameters
    ----------
    pid : int
        The process id of the launched instance
    """

    def __init__(self, proc: subprocess.Popen) -> None:
        self._proc: subprocess.Popen = proc
        self._returncode: Optional[int] = None
        self._rendering = False
        self._lines_read = 0
        self._scanner_thread = threading.Thread(
            target=OmniverseKitInstance._scan_stdout, args=(self,)
        )
        self._scanner_thread.start()
        self._simba_url: Optional[ParseResult] = None

    def __del__(self) -> None:
        """Close down the instance on delete"""
        self.close()

    def close(self) -> None:
        """Shutdown the Omniverse instance

        If the instance associated with this object is still running,
        shut it down.
        """
        if not self.is_running():
            return
        proc = psutil.Process(self._proc.pid)
        for child in proc.children(recursive=True):
            if psutil.pid_exists(child.pid):
                # This can be a race condition, so it is ok if the child is dead already
                try:
                    child.terminate()
                except psutil.NoSuchProcess:
                    pass
        # Same issue, this process might already be shutting down, so NoSuchProcess is ok.
        try:
            proc.terminate()
        except psutil.NoSuchProcess:
            pass
        self._scanner_thread.join()

        # On a forced close, set a return code of 0
        self._returncode = 0
        self._simba_url = None

    @staticmethod
    def _scan_stdout(oki: "OmniverseKitInstance"):
        while oki._proc and oki._proc.poll() is None:
            if oki._proc.stdout is not None:
                output_line = oki._proc.stdout.readline().decode("utf-8")
                oki._lines_read = oki._lines_read + 1
                if "RTX ready" in output_line:
                    oki._rendering = True
                if output_line.startswith("Server running on "):
                    urlstr = output_line.removeprefix("Server running on ")
                    oki._simba_url = urlparse(urlstr)

    def is_rendering(self) -> bool:
        """Check if the instance has finished launching and is ready to render

        Returns
        -------
        bool
            True if the instance is ready to render.
        """
        return self.is_running() and self._rendering

    def is_running(self) -> bool:
        """Check if the instance is still running

        Returns
        -------
        bool
            True if the instance is still running.
        """
        if self._proc is None:
            return False
        if self._proc.poll() is None:
            return True
        return False

    def simba_url(self) -> Optional[ParseResult]:
        """Return the URL for the Simba server

        Returns
        -------
        ParseResult or None
            URL for the Simba server, or None if it is not running
        """
        return self._simba_url

    def returncode(self) -> Optional[int]:
        """Get the return code if the process has stopped, or None if still running

        Returns
        -------
        int or None
            Get the return code if the process has stopped, or None if still running
        """
        if self._returncode is not None:
            return self._returncode
        if self.is_running():
            return None
        self._returncode = self._proc.returncode
        return self._returncode


# Deprecated
def find_kit_filename(fallback_directory: Optional[str] = None) -> Optional[str]:
    """
    Use a combination of the current omniverse application and the information
    in the local .nvidia-omniverse/config/omniverse.toml file to come up with
    the pathname of a kit executable suitable for hosting another copy of the
    ansys.geometry.server kit.

    Returns
    -------
    Optional[str]
        The pathname of a kit executable or None

    """
    # parse the toml config file for the location of the installed apps
    try:
        import tomllib
    except ModuleNotFoundError:
        import pip._vendor.tomli as tomllib

    homedir = os.path.expanduser("~")
    ov_config = os.path.join(homedir, ".nvidia-omniverse", "config", "omniverse.toml")
    if not os.path.exists(ov_config):
        return None
    # read the Omniverse configuration toml file
    with open(ov_config, "r") as ov_file:
        ov_data = ov_file.read()
    config = tomllib.loads(ov_data)
    appdir = config.get("paths", {}).get("library_root", fallback_directory)

    # If we are running inside an Omniverse app, use that information
    try:
        import omni.kit.app

        # get the current application
        app = omni.kit.app.get_app()
        app_name = app.get_app_filename().split(".")[-1]
        app_version = app.get_app_version().split("-")[0]
        # and where it is installed
        appdir = os.path.join(appdir, f"{app_name}-{app_version}")
    except ModuleNotFoundError:
        # Names should be like: "C:\\Users\\foo\\AppData\\Local\\ov\\pkg\\create-2023.2.3\\launcher.toml"
        target = None
        target_version = None
        for d in glob.glob(os.path.join(appdir, "*", "launcher.toml")):
            test_dir = os.path.dirname(d)
            # the name will be something like "create-2023.2.3"
            name = os.path.basename(test_dir).split("-")
            if len(name) != 2:
                continue
            if name[0] not in ("kit", "create", "view"):
                continue
            if (target_version is None) or (name[1] > target_version):
                target = test_dir
                target_version = name[1]
        if target is None:
            return None
        appdir = target

    # Windows: 'kit.bat' in '.' or 'kit' followed by 'kit.exe' in '.' or 'kit'
    # Linux: 'kit.sh' in '.' or 'kit' followed by 'kit' in '.' or 'kit'
    exe_names = ["kit.sh", "kit"]
    if sys.platform.startswith("win"):
        exe_names = ["kit.bat", "kit.exe"]

    # look in 4 places...
    for dir_name in [appdir, os.path.join(appdir, "kit")]:
        for exe_name in exe_names:
            if os.path.exists(os.path.join(dir_name, exe_name)):
                return os.path.join(dir_name, exe_name)

    return None


# Deprecated
def launch_kit_instance(
    kit_path: Optional[str] = None,
    extension_paths: Optional[List[str]] = None,
    extensions: Optional[List[str]] = None,
    cli_options: Optional[List[str]] = None,
    log_file: Optional[str] = None,
    log_level: str = "warn",
) -> "OmniverseKitInstance":
    """Launch an Omniverse application instance

    Parameters
    ----------
    kit_path : Optional[str]
        The full pathname of to a binary capable of serving as a kit runner.
    extension_paths : Optional[List[str]]
        List of directory names to include the in search for kits.
    extensions : Optional[List[str]]
        List of kit extensions to be loaded into the launched kit instance.
    log_file : Optional[str]
        The name of a text file where the logging information for the instance will be saved.
    log_level : str
        The level of the logging information to record: "verbose", "info", "warn", "error", "fatal",
        the default is "warn".

    Returns
    -------
    OmniverseKitInstance
        The object interface for the launched instance

    Examples
    --------
    Run a simple, empty GUI kit instance.

    >>> from ansys.pyensight.core.utils import omniverse
    >>> ov = omniverse.launch_kit_instance(extensions=['omni.kit.uiapp'])

    """
    # build the command line
    if not kit_path:
        kit_path = find_kit_filename()
    if not kit_path:
        raise RuntimeError("Unable to find a suitable Omniverse kit install")
    cmd = [kit_path]
    if extension_paths:
        for path in extension_paths:
            cmd.extend(["--ext-folder", path])
    if extensions:
        for ext in extensions:
            cmd.extend(["--enable", ext])
    if cli_options:
        for opt in cli_options:
            cmd.append(opt)
    if log_level not in ("verbose", "info", "warn", "error", "fatal"):
        raise RuntimeError(f"Invalid logging level: {log_level}")
    cmd.append(f"--/log/level={log_level}")
    if log_file:
        cmd.append(f"--/log/file={log_file}")
        cmd.append("--/log/enabled=true")
    # Launch the process
    env_vars = os.environ.copy()
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env_vars)
    return OmniverseKitInstance(p)


def find_app(ansys_installation: Optional[str] = None) -> Optional[str]:
    dirs_to_check = []
    if ansys_installation:
        # Given a different Ansys install
        local_tp = os.path.join(os.path.join(ansys_installation, "tp", "showcase"))
        if os.path.exists(local_tp):
            dirs_to_check.append(local_tp)
        # Dev Folder
        omni_platform_dir = "linux-x86_64"
        if sys.platform.startswith("win"):
            omni_platform_dir = "windows-x86_64"
        local_dev_omni = os.path.join(
            ansys_installation,
            "omni_build",
            "kit-app-template",
            "_build",
            omni_platform_dir,
            "release",
        )
        if os.path.exists(local_dev_omni):
            dirs_to_check.append(local_dev_omni)
    if "PYENSIGHT_ANSYS_INSTALLATION" in os.environ:
        env_inst = os.environ["PYENSIGHT_ANSYS_INSTALLATION"]
        dirs_to_check.append(os.path.join(env_inst, "tp", "showcase"))

    # Look for most recent Ansys install, 25.2 or later
    awp_roots = []
    for env_name in dict(os.environ).keys():
        if env_name.startswith("AWP_ROOT") and int(env_name[len("AWP_ROOT") :]) >= 252:
            awp_roots.append(env_name)
    awp_roots.sort(reverse=True)
    for env_name in awp_roots:
        dirs_to_check.append(os.path.join(os.environ[env_name], "tp", "showcase"))

    # check all the collected locations in order
    for install_dir in dirs_to_check:
        launch_file = os.path.join(install_dir, "ansys_tools_omni_core.py")
        if os.path.isfile(launch_file):
            return launch_file
    return None


def launch_app(
    usd_file: Optional[str] = "",
    layout: Optional[str] = "default",
    streaming: Optional[bool] = False,
    offscreen: Optional[bool] = False,
    log_file: Optional[str] = None,
    log_level: Optional[str] = "warn",
    cli_options: Optional[List[str]] = None,
    ansys_installation: Optional[str] = None,
    interpreter: Optional[str] = None,
) -> "OmniverseKitInstance":
    """Launch the Ansys Omniverse application

    Parameters
    ----------
    # usd_file : Optional[str]
    #    A .usd file to open on startup
    # layout : Optional[str]
    #    A UI layout.  viewer, composer, or composer_slim
    # streaming : Optional[bool]
    #    Enable webrtc streaming to enable the window in a web page
    # offscreen : Optional[str]
    #    Run the app offscreen.  Useful when streaming.
    # log_file : Optional[str]
    #    The name of a text file where the logging information for the instance will be saved.
    # log_level : Optional[str]
    #    The level of the logging information to record: "verbose", "info", "warn", "error", "fatal",
    #    the default is "warn".
    # cli_options : Optional[List[str]]
    #    Other command line options

    Returns
    -------
    OmniverseKitInstance
        The object interface for the launched instance

    Examples
    --------
    Run the app with default options

    >>> from ansys.pyensight.core.utils import omniverse
    >>> ov = omniverse.launch_app()

    """
    cmd = [sys.executable]
    if interpreter:
        cmd = [interpreter]
    app = find_app(ansys_installation=ansys_installation)
    if not app:
        raise RuntimeError("Unable to find the Ansys Omniverse app")
    cmd.extend([app])
    if usd_file:
        cmd.extend(["-f", usd_file])
    if layout:
        cmd.extend(["-l", layout])
    if streaming:
        cmd.extend(["-s"])
    if offscreen:
        cmd.extend(["-o"])
    if cli_options:
        cmd.extend(cli_options)
    if log_level:
        if log_level not in ("verbose", "info", "warn", "error", "fatal"):
            raise RuntimeError(f"Invalid logging level: {log_level}")
        cmd.extend([f"--/log/level={log_level}"])
    if log_file:
        cmd.extend(["--/log/enabled=true", f"--/log/file={log_file}"])

    # Launch the process
    env_vars = os.environ.copy()
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env_vars)
    return OmniverseKitInstance(p)


class Omniverse:
    """Provides the ``ensight.utils.omniverse`` interface.

    The omniverse class methods provide an interface between an EnSight session
    and an Omniverse instance.  See :ref:`omniverse_info` for additional details.

    Parameters
    ----------
    interface: Union["ensight_api.ensight", "ensight"]
        Entity that provides the ``ensight`` namespace. In the case of
        EnSight Python, the ``ensight`` module is passed. In the case
        of PyEnSight, ``Session.ensight`` is passed.

    Notes
    -----
    This interface is only available when using pyensight (they do not work with
    the ensight Python interpreter) and the module must be used in an interpreter
    that includes the Omniverse Python modules (e.g. omni and pxr).  Only a single
    Omniverse connection can be established within a single pyensight session.

    Examples
    --------

    >>> from ansys.pyensight.core import LocalLauncher
    >>> session = LocalLauncher().start()
    >>> ov = session.ensight.utils.omniverse
    >>> ov.create_connection("D:\\Omniverse\\Example")
    >>> ov.update()
    >>> ov.close_connection()

    """

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface
        self._server_pid: Optional[int] = None
        self._interpreter: str = ""
        self._status_filename: str = ""

    def _check_modules(self) -> None:
        """Verify that the Python interpreter is correct

        Check for module dependencies. If not present, raise an exception.

        Raises
        ------
        RuntimeError
            if the necessary modules are missing.

        """
        # One time check for this
        if len(self._interpreter):
            return

        # if a module, then we are inside EnSight
        if isinstance(self._ensight, ModuleType):  # pragma: no cover
            # in this case, we can just use cpython
            import ceiversion
            import enve

            cei_home = os.environ.get("CEI_HOME", enve.home())
            self._interpreter = os.path.join(cei_home, "bin", f"cpython{ceiversion.apex_suffix}")
            if platform.system() == "Windows":
                self._interpreter += ".bat"
            return
        # Check if the python interpreter is kit itself
        is_omni = False
        try:
            import omni  # noqa: F401

            is_omni = "kit" in os.path.basename(sys.executable)
        except ModuleNotFoundError:
            pass
        # Using the python interpreter running this code
        self._interpreter = sys.executable
        if "fluids_one" in self._interpreter:  # compiled simba-app
            self._interpreter = _handle_fluids_one(self._ensight._session._install_path)
        if is_omni:
            kit_path = os.path.dirname(sys.executable)
            self._interpreter = os.path.join(kit_path, "python")
            if platform.system() == "Windows":
                self._interpreter += ".bat"
            else:
                self._interpreter += ".sh"

        # in the future, these will be part of the pyensight wheel
        # dependencies, but for now we include this check.
        try:
            import pxr  # noqa: F401
            import pygltflib  # noqa: F401
        except Exception:
            raise RuntimeError("Unable to detect omniverse dependencies: usd-core, pygltflib.")

    def is_running_omniverse(self) -> bool:
        """Check that an Omniverse connection is active

        Returns
        -------
        bool
            True if the connection is active, False otherwise.
        """
        if self._server_pid is None:
            return False
        if psutil.pid_exists(self._server_pid):
            return True
        self._server_pid = None
        return False

    def create_connection(
        self,
        omniverse_path: str,
        include_camera: bool = False,
        normalize_geometry: bool = False,
        temporal: bool = False,
        live: bool = True,
        debug_filename: str = "",
        time_scale: float = 1.0,
        line_width: float = 0.0,
        options: dict = {},
    ) -> None:
        """Ensure that an EnSight dsg -> omniverse server is running

        Connect the current EnSight session to an Omniverse server.
        This is done by launching a new service that makes a dynamic scene graph
        connection to the EnSight session and pushes updates to the Omniverse server.
        The initial EnSight scene will be pushed after the connection is established.

        Parameters
        ----------
        omniverse_path : str
            The directory name where the USD files should be saved. For example:
            "C:/Users/test/OV/usdfiles"
        include_camera : bool
            If True, apply the EnSight camera to the Omniverse scene.  This option
            should be used if the target viewer is in AR/VR mode.  Defaults to False.
        normalize_geometry : bool
            Omniverse units are in meters.  If the source dataset is not in the correct
            unit system or is just too large/small, this option will remap the geometry
            to a unit cube.  Defaults to False.
        temporal : bool
            If True, save all timesteps.
        live : bool
            If True, one can call 'update()' to send updated geometry to Omniverse.
            If False, the Omniverse connection will push a single update and then
            disconnect.  Defaults to True.
        time_scale : float
            Multiply all EnSight time values by this factor before exporting to Omniverse.
            The default is 1.0.
        debug_filename : str
            If the name of a file is provided, it will be used to save logging information on
            the connection between EnSight and Omniverse.  This option is no longer supported,
            but the API remains for backwards compatibility.
        line_width : float
            If set, line objects will be represented as "tubes" of the size specified by
            this factor.  The default is 0.0 and causes lines not to be exported.
        options : dict
            Allows for a fallback for the grpc host/port and the security token.
        """
        if not isinstance(self._ensight, ModuleType):
            self._ensight._session.ensight_version_check("2023 R2")
        self._check_modules()
        if self.is_running_omniverse():
            raise RuntimeError("An Omniverse server connection is already active.")
        dsg_uri = None
        is_win = "Win" in platform.system()
        grpc_use_tcp_sockets = False
        grpc_allow_network_connectsion = False
        grpc_disable_tls = False
        if not isinstance(self._ensight, ModuleType):
            # Make sure the internal ui module is loaded
            grpc_use_tcp_sockets = self._ensight._session._grpc_use_tcp_sockets
            grpc_allow_network_connectsion = self._ensight._session._grpc_allow_network_connections
            grpc_disable_tls = self._ensight._session._grpc_disable_tls
            self._ensight._session.cmd("import enspyqtgui_int", do_eval=False)
            # Get the gRPC connection details and use them to launch the service
            use_tcp_sockets = self._ensight._session._grpc_use_tcp_sockets
            hostname = self._ensight._session.grpc.host
            token = self._ensight._session.grpc.security_token
            if not is_win and not use_tcp_sockets:
                uds_path = self._ensight._session._grpc_uds_pathname
                dsg_uds_path = "/tmp/greeter"
                if uds_path:
                    dsg_uds_path = uds_path
                dsg_uri = f"unix:{dsg_uds_path}.sock"
            else:
                port = self._ensight._session._grpc_port
                hostname = self._ensight._session.grpc.host
                token = self._ensight._session.grpc.security_token
                dsg_uri = f"grpc://{hostname}:{port}"
        else:
            hostname = options.get("host", "127.0.0.1")
            port = options.get("port", 12345)
            uds_path = options.get("uds_path")
            token = options.get("security", "")
            if uds_path and not is_win:
                dsg_uri = f"unix:{uds_path}.sock"
            else:
                dsg_uri = f"grpc://{hostname}:{port}"

        # Launch the server via the 'ansys.pyensight.core.utils.omniverse_cli' module
        cmd = [self._interpreter]
        cmd.extend(["-m", "ansys.pyensight.core.utils.omniverse_cli"])
        cmd.append(omniverse_path)
        if token:
            cmd.extend(["--security_token", token])
        if temporal:
            cmd.extend(["--temporal", "true"])
        if not include_camera:
            cmd.extend(["--include_camera", "false"])
        if normalize_geometry:
            cmd.extend(["--normalize_geometry", "true"])
        if time_scale != 1.0:
            cmd.extend(["--time_scale", str(time_scale)])
        if line_width != 0.0:
            cmd.extend(["--line_width", str(line_width)])
        if not live:
            cmd.extend(["--oneshot", "1"])
        if grpc_allow_network_connectsion:
            cmd.extend(["--grpc_allow_network_connections", "1"])
        if grpc_disable_tls:
            cmd.extend(["--grpc_disable_tls", "1"])
        if grpc_use_tcp_sockets:
            cmd.extend(["--grpc_use_tcp_sockets", "1"])
        cmd.extend(["--dsg_uri", dsg_uri])
        env_vars = os.environ.copy()
        # we are launching the kit from EnSight or PyEnSight.  In these cases, we
        # inform the kit instance of:
        # (1) the name of the "server status" file, if any
        self._new_status_file()
        env_vars["ANSYS_OV_SERVER_STATUS_FILENAME"] = self._status_filename
        process = subprocess.Popen(cmd, close_fds=True, env=env_vars)
        self._server_pid = process.pid

    def _new_status_file(self, new=True) -> None:
        """
        Remove any existing status file and create a new one if requested.

        Parameters
        ----------
        new : bool
            If True, create a new status file.
        """
        if self._status_filename:
            try:
                os.remove(self._status_filename)
            except OSError:
                pass
        self._status_filename = ""
        if new:
            self._status_filename = os.path.join(
                tempfile.gettempdir(), str(uuid.uuid1()) + "_gs_status.txt"
            )

    def read_status_file(self) -> dict:
        """Read the status file and return its contents as a dictionary.

        Note: this can fail if the file is being written to when this call is made, so expect
        failures.

        Returns
        -------
        Optional[dict]
            A dictionary with the fields 'status', 'start_time', 'processed_buffers', 'total_buffers' or empty
        """
        if not self._status_filename:
            return {}
        try:
            with open(self._status_filename, "r") as status_file:
                data = json.load(status_file)
        except Exception:
            return {}
        return data

    def close_connection(self) -> None:
        """Shut down the open EnSight dsg -> omniverse server

        Break the connection between the EnSight instance and Omniverse.

        """
        self._check_modules()
        if not self.is_running_omniverse():
            return
        proc = psutil.Process(self._server_pid)
        for child in proc.children(recursive=True):
            if psutil.pid_exists(child.pid):
                # This can be a race condition, so it is ok if the child is dead already
                try:
                    child.kill()
                except psutil.NoSuchProcess:
                    pass
        # Same issue, this process might already be shutting down, so NoSuchProcess is ok.
        try:
            proc.kill()
        except psutil.NoSuchProcess:
            pass
        self._server_pid = None
        self._new_status_file(new=False)

    def update(self, temporal: bool = False, line_width: float = 0.0) -> None:
        """Update the geometry in Omniverse

        Export the current EnSight scene to the current Omniverse connection.

        Parameters
        ----------
        temporal : bool
            If True, export all timesteps.
        line_width : float
            If set to a non-zero value, lines will be exported with this thickness.
            This feature is only available in 2025 R2 and later.
        """
        update_cmd = "dynamicscenegraph://localhost/client/update"
        prefix = "?"
        if temporal:
            update_cmd += f"{prefix}timesteps=1"
            prefix = "&"
        if line_width != 0.0:
            add_linewidth = False
            if isinstance(self._ensight, ModuleType):
                add_linewidth = True
            else:
                # only in 2025 R2 and beyond
                if self._ensight._session.ensight_version_check("2025 R2", exception=False):
                    add_linewidth = True
            if add_linewidth:
                update_cmd += f"{prefix}ANSYS_linewidth={line_width}"
                prefix = "&"
        self._check_modules()
        if not self.is_running_omniverse():
            raise RuntimeError("No Omniverse server connection is currently active.")
        if not isinstance(self._ensight, ModuleType):
            self._ensight._session.ensight_version_check("2023 R2")
            cmd = f'enspyqtgui_int.dynamic_scene_graph_command("{update_cmd}")'
            self._ensight._session.cmd(cmd, do_eval=False)
        else:
            import enspyqtgui_int

            enspyqtgui_int.dynamic_scene_graph_command(f"{update_cmd}")



""" This module provides a list of common utilities shared between different PyEnSight modules."""

import os
import random
import re
import socket
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from ansys.pyensight.core import enshell_grpc
import urllib3

try:
    from simple_upload_server.client import Client

    simple_upload_server_is_available = True  # pragma: no cover
except Exception:
    simple_upload_server_is_available = False

if TYPE_CHECKING:
    from docker import DockerClient

GRPC_VERSIONS = ["2025 R2.3", "2025 R1.4", "2024 R2.5"]
GRPC_WARNING_MESSAGE = "The EnSight version being used uses an insecure gRPC connection."
GRPC_WARNING_MESSAGE += "Consider upgrading to a version higher than 2026 R1, or "
GRPC_WARNING_MESSAGE += "2024 R2.5 or higher service packs, or 2025 R1.4 or higher service packs "
GRPC_WARNING_MESSAGE += "or 2025 R2.3 or higher service packs."


def find_unused_ports(count: int, avoid: Optional[List[int]] = None) -> Optional[List[int]]:
    """Find "count" unused ports on the host system

    A port is considered unused if it does not respond to a "connect" attempt.  Walk
    the ports from 'start' to 'end' looking for unused ports and avoiding any ports
    in the 'avoid' list.  Stop once the desired number of ports have been
    found.  If an insufficient number of ports were found, return None.

    Parameters
    ----------
    count: int :
        Number of unused ports to find
    avoid: Optional[List[int]] :
        An optional list of ports not to check

    Returns
    -------
        The detected ports or None on failure

    """
    if avoid is None:
        avoid = []
    ports = list()

    # pick a starting port number
    start = random.randint(1024, 64000)
    # We will scan for 65530 ports unless end is specified
    port_mod = 65530
    end = start + port_mod - 1
    # walk the "virtual" port range
    for base_port in range(start, end + 1):
        # Map to physical port range
        # There have been some issues with 65534+ so we stop at 65530
        port = base_port % port_mod
        # port 0 is special
        if port == 0:  # pragma: no cover
            continue  # pragma: no cover
        # avoid admin ports
        if port < 1024:  # pragma: no cover
            continue  # pragma: no cover
        # are we supposed to skip this one?
        if port in avoid:  # pragma: no cover
            continue  # pragma: no cover
        # is anyone listening?
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex(("127.0.0.1", port))
        if result != 0:
            ports.append(port)
        else:
            sock.close()  # pragma: no cover
        if len(ports) >= count:
            return ports
    # in case we failed...
    if len(ports) < count:  # pragma: no cover
        return None  # pragma: no cover
    return ports  # pragma: no cover


def get_host_port(uri: str) -> Tuple[str, int]:
    """Get the host port for the input uri

    Parameters
    ----------

    uri: str
        The Uri to inspect

    Returns
    -------
    (tuple):
        A tuple containing the host and the port of the input uri
    """
    parse_results = urllib3.util.parse_url(uri)
    port = (
        parse_results.port
        if parse_results.port
        else (443 if re.search(r"^https|wss$", parse_results.scheme) else None)
    )
    return (parse_results.host, port)


def get_file_service(pim_instance: Any) -> Optional[Any]:  # pragma: no cover
    """Get the file service object for the input pim instance.

    Parameters
    ----------

    pim_instance:
        the PIM instance to get the service from.

    Returns
    -------

    pim_file_service:
        the PIM file service object
    """
    if simple_upload_server_is_available is False:
        return None
    if pim_instance is None:
        return None

    if "http-simple-upload-server" in pim_instance.services:
        pim_file_service = Client(
            token="token",
            url=pim_instance.services["http-simple-upload-server"].uri,
            headers=pim_instance.services["http-simple-upload-server"].headers,
        )
        return pim_file_service
    return None


def populate_service_host_port(  # pragma: no cover
    pim_instance: Any, service_host_port: Dict[str, Tuple[str, int]], webui: bool = False
) -> Dict[str, Tuple[str, int]]:
    """Populate the service host port dictionary with the services available in the PIM instance.

    Parameters
    ----------
    pim_instance:
        the PIM instance to get the servicea from.
    service_host_port: dict
        the dictionary to be updated with the services from the PIM instance
    webui: bool
        if True retrieve also the webUI service

    Returns
    -------
    service_host_port: dict
        the dictionary updated with the services from the PIM instance
    """
    if not set(("grpc_private", "http", "ws")).issubset(pim_instance.services):
        raise RuntimeError(
            "If channel is specified, the PIM instance must have a list of length 3 "
            + "containing the appropriate service URIs. It does not."
        )
    service_host_port["grpc_private"] = get_host_port(pim_instance.services["grpc_private"].uri)
    service_host_port["http"] = get_host_port(pim_instance.services["http"].uri)
    service_host_port["ws"] = get_host_port(pim_instance.services["ws"].uri)
    service_host_port["grpc"] = ("127.0.0.1", -1)
    if webui:
        service_host_port["webui"] = get_host_port(pim_instance.services["webui"].uri)
    return service_host_port


def launch_enshell_interface(
    enshell_grpc_channel: Any,
    grpc_port: int,
    timeout: float,
    secret_key: str,
    grpc_use_tcp_sockets: bool = False,
    grpc_allow_network_connections: bool = False,
    grpc_disable_tls: bool = False,
    grpc_uds_pathname: Optional[str] = None,
) -> enshell_grpc.EnShellGRPC:
    """Launch the EnShell gRPC Interface.

    Parameters
    ----------
    enshell_grpc_channel:
        An eventual gRPC channel already available, like in the PIM case
    grpc_port: int
        the gRPC port to connect to
    timeout: float
        a timeout to wait for the gRPC connection
    grpc_use_tcp_sockets: bool
        If using gRPC, and if True, then allow TCP Socket based connections
        instead of only local connections.
    grpc_allow_network_connections: bool
        If using gRPC and using TCP Socket based connections, listen on all networks.
    grpc_disable_tls: bool
        If using gRPC and using TCP Socket based connections, disable TLS.
    grpc_uds_pathname: bool
        If using gRPC and using Unix Domain Socket based connections, explicitly
        set the pathname to the shared UDS file instead of using the default.

    Returns
    -------
    enshell: enshell_grpc.EnShellGRPC
        the enshell gRPC interface

    WARNING:
    Overriding the default values for these options: grpc_use_tcp_sockets, grpc_allow_network_connections,
    and grpc_disable_tls
    can possibly permit control of this computer and any data which resides on it.
    Modification of this configuration is not recommended.  Please see the
    documentation for your installed product for additional information.
    """
    if enshell_grpc_channel:  # pragma: no cover
        enshell = enshell_grpc.EnShellGRPC(
            grpc_use_tcp_sockets=grpc_use_tcp_sockets,
            grpc_allow_network_connections=grpc_allow_network_connections,
            grpc_disable_tls=grpc_disable_tls,
            grpc_uds_pathname=grpc_uds_pathname,
        )  # pragma: no cover
        enshell.security_token = secret_key
        enshell.connect_existing_channel(enshell_grpc_channel)  # pragma: no cover
    else:
        enshell = enshell_grpc.EnShellGRPC(
            port=grpc_port,
            grpc_use_tcp_sockets=grpc_use_tcp_sockets,
            grpc_allow_network_connections=grpc_allow_network_connections,
            grpc_disable_tls=grpc_disable_tls,
            grpc_uds_pathname=grpc_uds_pathname,
        )
        enshell.security_token = secret_key
        time_start = time.time()
        while time.time() - time_start < timeout:  # pragma: no cover
            if enshell.is_connected():
                break
            try:
                enshell.connect(timeout=timeout)
            except OSError:  # pragma: no cover
                pass  # pragma: no cover
    return enshell


def pull_image(docker_client: "DockerClient", image_name: str) -> None:
    """Pull the input docker image using the input Docker Client

    Parameters
    ----------
    docker_client: DockerClient
        the current DockerClient to pull the image with
    image_name: str
        the image to pull
    """
    try:
        if docker_client is not None:  # pragma: no cover
            docker_client.images.pull(image_name)
    except Exception:
        raise RuntimeError(f"Can't pull Docker image: {image_name}")


def _is_within_directory(directory, target):
    """Check if target is inside of the input directory."""
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)
    return os.path.commonprefix([abs_directory, abs_target]) == abs_directory


def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
    """Utility to check tar extraction to avoid bandit check issue."""
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not _is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")
    tar.extractall(path, members, numeric_owner=numeric_owner)


def grpc_version_check(internal_version, ensight_full_version):
    """Check if the gRPC security options apply to the EnSight install."""
    if int(internal_version) < 242:
        return False
    # From 261 onward always available
    if int(internal_version) >= 261:
        return True
    # Exactly matches one of the service pack that introduced the changes
    if ensight_full_version in GRPC_VERSIONS:
        return True
    split_version = ensight_full_version.split(" ")
    year = split_version[0]
    # Not a service pack
    if len(split_version[1].split(".")) == 1:
        return False
    r_version = split_version[1].split(".")[0]
    dot_version = int(split_version[1].split(".")[1])
    # Check service pack version
    if year == "2024" and dot_version < 5:
        return False
    if year == "2025":
        if r_version == "R1" and dot_version < 4:
            return False
        if r_version == "R2" and dot_version < 3:
            return False
    return True



"""ensobj module

The ensobj module provides the base class to all EnSight proxy objects

"""
from typing import TYPE_CHECKING, Any, Optional, no_type_check

if TYPE_CHECKING:
    from ansys.pyensight.core import Session


class ENSOBJ(object):
    """Bass class for all EnSight proxy objects

    The ENSOBJ class is the base class for the EnSight object proxy interface.
    Note: this interface is internal and applications should not attempt to
    create ENSOBJ instances directly.

    Parameters
    ----------
    session :
        The session object associated with this object instance.
    objid :
        The EnSight CvfObjID of the object instance that this instance will
        serve as the proxy for.
    attr_id :
        For subclasses that differentiate classes by an attribute value,
        this is the attribute ID to use as the differentiator.  Example
        classes are ENS_TOOL, ENS_PART and ENS_ANNOT.
    attr_value :
        The attribute value associated with any specified attr_id.
    owned : bool
        If True, the object is assumed to be "owned" by this interpreter.
        This means that the lifecycle of the ENSOBJ instance in EnSight is
        dictated by the lifecycle of this proxy object.

    """

    def __init__(
        self,
        session: "Session",
        objid: int,
        attr_id: Optional[int] = None,
        attr_value: Optional[int] = None,
        owned: Optional[bool] = None,
    ) -> None:
        self._session = session
        self._objid = objid
        self._attr_id = attr_id
        self._attr_value = attr_value

        # True if this Python instance "owns" the ENSOBJ instance (via EnSight proxy cache)
        if owned:
            self._is_owned = True
        else:
            self._is_owned = False
            # do not put this object in the cache if it is owned, allow gc
            self._session.add_ensobj_instance(self)

    def __eq__(self, obj):
        return self._objid == obj._objid

    def __lt__(self, obj):
        return self._objid < obj._objid

    def __hash__(self):
        return self._objid

    def __del__(self):
        # release the session to allow for garbage collection
        tmp_session = self._session
        self._session = None
        if self._is_owned:
            try:
                cmd = f"ensight.objs.release_id('{tmp_session.name}', {self.__OBJID__})"
                tmp_session.cmd(cmd, do_eval=False)
            except Exception:  # pragma: no cover
                # This could happen at any time, including outside
                # the scope of the session, so we need to be
                # ready for any error.
                pass

    @property
    def __OBJID__(self) -> int:  # noqa: N802
        return self._objid

    def _remote_obj(self) -> str:
        """Convert the object into a string appropriate for use in the
        remote EnSight session.   Usually, this is some form of
        ensight.objs.wrap_id()."""
        return self._session.remote_obj(self._objid)

    def getattr(self, attrid: Any) -> Any:
        """Query the value of the specified attribute

        Parameters
        ----------
        attrid: Any
            The attribute to query

        Returns
        -------
        Any
            The current value of the attribute.

        Examples
        --------
        These commands are equivalent

        >>> v = part.VISIBLE
        >>> v = part.getattr("VISIBLE")
        >>> v = part.getattr(session.ensight.objs.enums.VISIBLE)
        """
        return self._session.cmd(f"{self._remote_obj()}.getattr({attrid.__repr__()})")

    def getattrs(self, attrid: Optional[list] = None, text: int = 0) -> dict:
        """Query the value of a collection of attributes

        This method queries a collection of attributes in a single call and
        returns the attribute values in a dictionary keyed by the attribute ids.

        Parameters
        ----------
        attrid : Optional[list]
            If this value is a list (of attribute ids as strings or enum values),
            then the returned dictionary will only include values for the specified
            attributes.  Otherwise, the returned dictionary will include values for
            all attributes of the target object.
        text : int
            By default, the returned dictionary keys are the attribute ID enum
            values.  If text is set to 1, the dictionary keys will be strings.
            Return:

        Returns
        -------
            Session CMD.

        Examples
        --------
        To copy some attributes from one part to another.

        >>> tmp = part0.getattrs(["VISIBLE", session.ensight.objs.enums.OPAQUENESS])
        >>> part1.setattrs(tmp)

        """
        if attrid is None:
            cmd = f"{self._remote_obj()}.getattrs(text={text})"
        else:
            cmd = f"{self._remote_obj()}.getattrs({attrid.__repr__()},text={text})"
        return self._session.cmd(cmd)

    def setattr(self, attrid: Any, value: Any) -> None:
        """Set the value of the specified attribute

        Parameters
        ----------
        attrid : Any
            The attribute to set.  This can be an integer (enum) or string.
        value : Any
            The value to set the attribute to.

        Returns
        -------
            Session CMD.

        Examples
        --------
        These commands are equivalent

        >>> part.setattr("VISIBLE", True)
        >>> part.getattr(session.ensight.objs.enums.VISIBLE, True)

        """
        return self._session.cmd(
            f"{self._remote_obj()}.setattr({attrid.__repr__()}, {value.__repr__()})"
        )

    def setattrs(self, values: dict, all_errors: int = 0) -> None:
        """Set the values of a collection of attributes

        Parameters
        ----------
        values : Dict
            The values to set.  The keys are the attribute IDs.
        all_errors : int
            If non-zero, raise a RuntimeError exception if any attribute set
            operation fails. By default, 0.

        Examples
        --------
        These commands are equivalent

        >>> part.VISIBLE = True
        >>> part.setattrs(dict(VISIBLE=True))
        >>> part.setattrs({session.ensight.objs.enums.VISIBLE: True})

        """
        cmd = f"{self._remote_obj()}.setattrs({values.__repr__()}, all_errors={all_errors})"
        return self._session.cmd(cmd)

    def attrinfo(self, attrid: Optional[Any] = None) -> dict:
        """For a given attribute id, return type information

        The dictionary returned by will always have the following keys:
            - `type` the high-level type. This can include any of the values noted in the next
              section

            - `basetype` the low-level type. Only types starting with `CVF` are allowed.

            - `numvals` the number of values of the `basetype`.

            - `desc` string description for the attr

            - `name` the python legal name for the attr

            - `flags` this integer is formed by or-ing values from the table below

        Optional keys:
            - `range` if selected in the flags, this key will be a list of two floats or
              ints: [min,max]. see `flags` below

            - `enums` if selected in the flags, this key will be list of lists. Each list has
              three values: [int_value, description, python_name]. The integer value is the number
              for the enum (see `flags` below), while the other two values are human and Python
              strings. The python_name will also be available as an integer in the
              ensight.objs.enums module.

            - `dependencies` if present, this key is a list of dictionaries that describe the
              fact that this attr is dependent on another attr. The dictionary is described below.

        Parameters
        ----------
        attrid : Optional[Any]
            The attribute to query

        Returns
        -------
        dict
            A dictionary that describes type information for the attribute.

        Examples
        --------
        >>> part.attrinfo(session.ensight.objs.enums.VISIBLE)

        """
        if not attrid:
            return self._session.cmd(f"{self._remote_obj()}.attrinfo()")
        return self._session.cmd(f"{self._remote_obj()}.attrinfo({attrid.__repr__()})")

    def attrissensitive(self, attrid: Any) -> bool:
        """Check to see if a given attribute is 'sensitive'

        Given the current state of the target object, return True if it makes sense to
        allow the specified attribute to change.   Return False if the object is currently
        ignoring the value of a given attribute due to the state of other attributes.

        Parameters
        ----------
        attrid : Any
            The attribute to query

        Returns
        -------
        bool
            True or False

        """
        return self._session.cmd(f"{self._remote_obj()}.attrissensitive({attrid.__repr__()})")

    def attrtree(
        self,
        all: int = 0,
        filter: Optional[list] = None,
        exclude: Optional[list] = None,
        include: Optional[list] = None,
        group_exclude: Optional[list] = None,
        group_include: Optional[list] = None,
        insensitive: int = 1,
    ) -> dict:
        """Get detailed GUI information for attributes of this object.

        This method is on MOST of the intrinsic objects, but not all of them. This method is used
        to generate a "visual" tree of an object's attributes. The most common use would be to
        dynamically generate labels and hierarchy for a "property sheet" editor.

        The method returns an object tree that describes the way attribute should be laid out.
        Each object has three attributes: `attr`, 'hasdeps' and `name`. All objects will have names
        and group objects will have an attr of -1. All objects can also be iterated for children
        objects of the same type. Only objects with an `attr` of -1 will have children. len() can
        be used to get the number of child objects. The top level object always has the name
        `root`. The 'hasdeps' attribute is the number of attributes in this attrtree() that have
        a dependency on this attr.  This can be used to decide if a new sensitivity check is needed
        if a given attribute changes.

        Parameters
        ----------
        all : int
            If set to 1 will include all attrs for the object, even if they are not in the group
            tables.
        filter : Optional[list]
            Should be set to an optional list of EnSight objects. The output will be filtered
            to include only the attributes in common between all of the objects (they do not
            need to be of the same type).
        include : Optional[list]
            Should be set to a list of attribute enums. Only the enums in the list will be
            included in the output object tree. Note: passing an empty list via this keyword,
            all the enums will be initially excluded.  This is useful with the
            group_include= keyword.
        exclude : Optional[list]
            Should be set to a list of attribute enums. Any enums in the list will be
            removed from the output object tree.
        group_exclude : Optional[list]
            Should be set to a list of attribute enums. For any enum in this list, exclude
            the enum and all the other enums in the same groups as the passed enums. Think
            of this as a shortcut for exclude= that requires you to only pass a single enum
            in the group you want to suppress.
        group_include : Optional[list]
            Should be set to a list of attribute enums. For any enum in this list, include
            the enum and all the other enums in the same groups as the passed enums. Think
            of this as a shortcut for include= that requires you to only pass a single enum
            in the group you want to include. Note: it may be necessary to pass include=[]
            (the empty list) to start with an empty list of enums.
        insensitive : int
            If this keyword is set to 0, attrtree() will call foo.issensitive() on each
            filtered attr and if the attr is not currently sensitive, it will remove it
            from the output.  The default value for this keyword is 1 which disables all
            sensitivity filtering.

        Examples
        --------

        >>> def walk_tree(part,obj,s):
        >>>     a = obj.attr
        >>>     if (a == -1):
        >>>         print("{}Group={}".format(s, obj.name)))
        >>>     else:
        >>>         = part.attrinfo(a)
        >>>         t = enum_to_name(a)
        >>>         d = info['desc']
        >>>         print("{}Attr={} - '{}' ({:d} deps)".format(s, t, d, obj.hasdeps))
        >>>     for i in obj:
        >>>         walk_tree(part,i,s+"  ")
        >>> walk_tree(session.ensight.core.PARTS[0],session.ensight.core.PARTS[0].attrtree(),"")
        """
        obj = f"{self._remote_obj()}"
        options = f"all={all}"
        options += f",insensitive={insensitive}"
        if filter:
            options += f",filter={filter.__repr__()}"
        if include:
            options += f",include={include.__repr__()}"
        if exclude:
            options += f",exclude={exclude.__repr__()}"
        if group_exclude:
            options += f",group_exclude={group_exclude.__repr__()}"
        if group_include:
            options += f",group_include={group_include.__repr__()}"
        return self._session.cmd(f"{obj}.attrgroupinfo({options})")

    def setattr_begin(self) -> None:
        """Signal bulk attribute update begin

        By default, EnSight will update the display with every attribute change.  If a lot of
        changes are to be made, this can slow execution down.  Putting a collection of changes
        inside of a setattr_begin()/setattr_end() pair of calls will defer the update until the
        setattr_end() call is made.

        """
        return self._session.cmd(f"{self._remote_obj()}.setattr_begin()")

    def setattr_end(self) -> None:
        """Signal bulk attribute update end

        By default, EnSight will update the display with every attribute change.  If a lot of
        changes are to be made, this can slow execution down.  Putting a collection of changes
        inside of a setattr_begin()/setattr_end() pair of calls will defer the update until the
        setattr_end() call is made.

        """
        return self._session.cmd(f"{self._remote_obj()}.setattr_end()")

    def setattr_status(self) -> int:
        """ """
        return self._session.cmd(f"{self._remote_obj()}.setattr_status()")

    def setmetatag(self, tag: str, value: Optional[Any]) -> None:
        """Change a value in the METADATA attribute

        All ENSOBJ subclasses have a METADATA read only attribute that is viewed as a Python
        dictionary.  A value can be set in that dictionary using this call:

        Parameters
        ----------
        tag : str
            The string name of the METADATA tag to add/change.
        value : Any, optional
            The value to change to tag to.  Note: this can be a string, int or float.

        Examples
        --------
        >>> session.ensight.objs.core.PARTS[0].setmetatag("FOO", "HELLO")
        >>> print(session.ensight.objs.core.PARTS[0].METADATA)

        """
        if value is None:
            cmd = f"{self._remote_obj()}.setmetatag({tag.__repr__()})"
        else:
            cmd = f"{self._remote_obj()}.setmetatag({tag.__repr__()}, {value.__repr__()})"
        return self._session.cmd(cmd)

    def hasmetatag(self, tag: str) -> bool:
        """Check to see if a tag exists in the METADATA attribute

        Parameters
        ----------
        tag : str
            The string name of the METADATA tag to check

        Returns
        -------
        bool
            True if the named tag exists in the METADATA attribute.

        """
        return self._session.cmd(f"{self._remote_obj()}.hasmetatag({tag.__repr__()})")

    def getmetatag(self, tag: str) -> Any:
        """Get the value of a tag in the METADATA attribute

        Parameters
        ----------
        tag : str

        The string name of the METADATA tag to get

        Returns
        -------
        Any
            The value assigned to the tag in the METADATA attribute.

        Examples
        --------
        >>> session.ensight.objs.core.PARTS[0].setmetatag("FOO", "HELLO")
        >>> print(session.ensight.objs.core.PARTS[0].getmetatag("FOO"))

        """
        return self._session.cmd(f"{self._remote_obj()}.getmetatag({tag.__repr__()})")

    def destroy(self) -> None:
        """Destroy the EnSight object associated with this proxy object"""
        self._session.cmd(f"{self._remote_obj()}.destroy()")

    def __str__(self) -> str:
        desc = ""
        if hasattr(self.__class__, "attr_list"):  # pragma: no cover
            if self._session.ensight.objs.enums.DESCRIPTION in self.__class__.attr_list:
                try:
                    if hasattr(self, "DESCRIPTION"):  # pragma: no cover
                        desc_text = self.DESCRIPTION
                    else:
                        desc_text = ""
                except RuntimeError:  # pragma: no cover
                    # self.DESCRIPTION is a gRPC call that can fail for default objects
                    desc_text = ""  # pragma: no cover
                desc = f", desc: '{desc_text}'"
        owned = ""
        if self._is_owned:
            owned = ", Owned"
        return f"Class: {self.__class__.__name__}{desc}{owned}, CvfObjID: {self._objid}, cached:no"

    def __repr__(self) -> str:
        """Custom __repr__ method used by the stub API.

        In some cases, we need to specify the object representation in the EnSight
        instance.  For ENSOBJ instances, this means using the wrap_id() mechanism
        instead of the built-in __repr__.  So the generated API uses this interface.
        """
        return f"ensight.objs.wrap_id({self._objid})"

    @no_type_check
    def _repr_pretty_(self, p: "pretty", cycle: bool) -> None:
        """Support the pretty module for better IPython support.

        Parameters
        ----------
        p: str
            pretty flag.

        cycle: bool :
            cycle flag.

        """
        p.text(self.__str__())
